from __future__ import absolute_import, division, print_function

from pandora.compat import text_type


class Undefined(object):
    pass


class Test(object):
    AND = '&'
    OR = '|'

    def __init__(self, commands, mode=AND, error_or='No test has passed.'):
        self.commands = commands
        self.mode = mode

        self.error_or = error_or

    def run(self, value):
        if self.mode == self.AND:
            return self.run_and(value)
        elif self.mode == self.OR:
            return self.run_or(value)
        else:
            raise RuntimeError('Invalid mode supplied.')

    def run_and(self, value):
        for command in self.commands:
            value = command(value)

        return value

    def run_or(self, value):
        for command in self.commands:
            try:
                return command(value)
            except ValueError:
                continue

        raise ValueError(self.error_or)

    def __call__(self, value):
        return self.run(value)


class MultiValueError(ValueError):
    def __init__(self, message, errors):
        super(MultiValueError, self).__init__(message)
        self.errors = errors


class Mapping(object):
    def __init__(self, schema):
        self.schema = list()

        for field_name in schema.keys():
            if not isinstance(field_name, Field):
                field = Field(field_name)
            else:
                field = field_name

            self.schema.append([field, schema[field_name]])

    def __call__(self, data, error='Validation error.',
                 error_required='Field is required.'):
        result = dict()
        errors = dict()

        for field, test in self.schema:
            field_name = str(field)

            if field_name not in data:
                if field.optional:
                    result[field_name] = field.default
                else:
                    errors[field_name] = error_required
            else:
                try:
                    result[field_name] = test(data[field_name])
                except MultiValueError as exc:
                    errors[field_name] = exc
                except ValueError as exc:
                    errors[field_name] = str(exc)

        if len(errors):
            raise MultiValueError(error, errors)

        return result


class Field(object):
    def __init__(self, name, default=Undefined):
        self.name = name
        self.optional = default is not Undefined
        self.default = default

    def __unicode__(self):
        return self.name

    def __lt__(self, other):
        return str(self) < str(other)

    def __str__(self):
        return text_type(self.name)


def validate(callback, error='Validation error.'):
    def wrapper(value):
        if not callback(value):
            raise ValueError(error)
        return value

    return wrapper


def test_and_return(callback, return_value, error='Validation error.'):
    def wrapper(value):
        if callback(value):
            return return_value

        raise ValueError('Validation error.')

    return wrapper



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from disinfect import Field
from disinfect import Mapping
from disinfect import MultiValueError
from disinfect import Test
from disinfect import tests
from disinfect import validate
from pandora.compat import text_type
from pytest import raises


class TestDisinfect(object):
    def test_basics(self):
        test = Test([
            lambda v: text_type(v),
            validate(lambda v: len(v) >= 1, error='Not long enough.'),
            validate(lambda v: len(v) <= 10, error='Too long.'),
        ])

        assert test('Nils') == u'Nils'

        with raises(ValueError) as exc:
            test('')

        assert exc.type is ValueError
        assert str(exc.value) == 'Not long enough.'

    def test_string(self):
        test = tests.string_field()
        test('Nils') == 'Nils'
        test(u'Níls') == u'Níls'

    def test_mapping(self):
        def string_field(min_len=1, max_len=150):
            return Test([
                lambda v: text_type(v),
                validate(lambda v: len(v) >= min_len,
                         error='Not long enough.'),
                validate(lambda v: len(v) <= max_len,
                         error='Too long.'),
            ])

        mapping = Mapping({
            'first': string_field(),
            Field('infix', default=''): string_field(min_len=0, max_len=40),
            'last': string_field(),
        })

        user = mapping({
            'first': 'Nils',
            'last': 'Corver',
        })

        assert user == {
            'first': 'Nils',
            'infix': '',
            'last': 'Corver',
        }

        with raises(MultiValueError) as exc:
            mapping({})

        assert exc.type is MultiValueError
        assert exc.value.errors == {
            'first': 'Field is required.',
            'last': 'Field is required.',
        }

    def test_tests(self):
        test = tests.boolean_field()

        assert test('f') is False
        assert test('true') is True

        with raises(ValueError) as exc:
            test('')

        assert str(exc.value) == 'Not a boolean value.'

    def test_complex(self):
        mapping = Mapping({
            'first': tests.string_field(),
            Field('infix', default=''): tests.string_field(min_len=0,
                                                           max_len=40),
            'last': tests.string_field(),

            'addresses': tests.list_of(Mapping({
                'zipcode': tests.string_field(min_len=6, max_len=6),
                'housenumber': tests.int_field(min_value=1),
            }))
        })

        user = mapping({
            'first': 'Nils',
            'infix': '',
            'last': 'Corver',
            'addresses': [
                {'zipcode': '2024HP', 'housenumber': '118'},
                {'zipcode': '2024VV', 'housenumber': '38'},
            ]
        })

        assert user == {
            'first': 'Nils',
            'infix': '',
            'last': 'Corver',
            'addresses': [
                {'zipcode': '2024HP', 'housenumber': 118},
                {'zipcode': '2024VV', 'housenumber': 38},
            ]
        }

    def test_enum(self):
        test = tests.enum_field(['red', 'green', 'blue'])
        assert test('red') == 'red'

        test = tests.enum_field([1, 2, 3, 4], tests.int_field())
        assert test('3') == 3

    def test_set(self):
        test = tests.set_field(['red', 'green', 'blue'])
        assert test('red,blue') == ['red', 'blue']

        test = tests.set_field([1, 2, 3, 4], tests.int_field())
        assert test('2,3') == [2, 3]



from . import Test
from . import validate
from pandora.compat import string_types
from pandora.compat import text_type
from . import test_and_return
import bleach

BOOL_FIELD_TRUE_VALUES = ['t', '1', 1, True, 'true', 'yes', 'True']
BOOL_FIELD_FALSE_VALUES = ['f', '-1', '0', -1, 0, False, 'false', 'no',
                           'False']
BOOL_FIELD_NONE_VALUES = ['null', 'none', None, 'None', 'undefined',
                          'Undefined']


def string_field(min_len=1, max_len=150,
                 min_error='Too short.', max_error='Too long.',
                 strip_html=True):
    commands = list()
    commands.append(lambda v: text_type(v))

    if min_len is not None:
        commands.append(validate(lambda v: len(v) >= min_len, min_error))
    if max_len is not None:
        commands.append(validate(lambda v: len(v) <= max_len, max_error))

    if strip_html:
        commands.append(lambda v: bleach.clean(v, tags=list(), strip=True))

    return Test(commands)


def int_field(min_value=None, max_value=None,
              min_error='Too low.', max_error='Too high.'):
    commands = list()
    commands.append(lambda v: int(v))

    if min_value is not None:
        commands.append(validate(lambda v: v >= min_value, min_error))
    if max_value is not None:
        commands.append(validate(lambda v: v <= max_value, max_error))

    return Test(commands)


def float_field(min_value=None, max_value=None,
                min_error='Too low.', max_error='Too high.'):
    commands = list()
    commands.append(lambda v: float(v))

    if min_value is not None:
        commands.append(validate(lambda v: v >= min_value, min_error))
    if max_value is not None:
        commands.append(validate(lambda v: v <= max_value, max_error))

    return Test(commands)


def boolean_field(false_values=None, true_values=None, true_unless_false=True,
                  none_values=None, error='Not a boolean value.'):
    if false_values is None:
        false_values = BOOL_FIELD_FALSE_VALUES
    if true_values is None:
        true_values = BOOL_FIELD_TRUE_VALUES

    commands = list()

    if none_values is not None:
        commands.append(test_and_return(lambda v: v in none_values, None))

    if true_values is None and true_unless_false:
        commands.append(lambda v: v not in false_values)
    else:
        commands.append(test_and_return(lambda v: v in false_values, False))
        commands.append(test_and_return(lambda v: v in true_values, True))

    return Test(commands, mode=Test.OR, error_or=error)


def list_field(sanitize=None, split_character=','):
    test = Test([
        Test([
            validate(lambda v: isinstance(v, string_types)),
            validate(lambda v: len(v) > 0),
            lambda v: v.split(split_character),
        ]),
        Test([
            validate(lambda v: isinstance(v, (list, set,))),
        ]),
    ], mode=Test.OR)

    if sanitize:
        return Test([
            test,
            lambda v: [sanitize(x) for x in v],
        ])

    return test


def enum_field(allowed_values, sanitize=None):
    if sanitize is None:
        sanitize = string_field()

    return Test([
        sanitize,
        validate(lambda v: v in allowed_values),
    ])


def set_field(allowed_values, sanitize=None, split_character=','):
    return Test([
        list_field(sanitize, split_character),
        validate(lambda v: [x in allowed_values for x in v]),
    ])


def list_of(test, sanitize=None, split_character=','):
    return Test([
        list_field(sanitize, split_character),
        lambda v: [test(x) for x in v],
    ])



from __future__ import absolute_import, division, print_function

from .disinfect import (
    Field,
    Mapping,
    MultiValueError,
    Test,
    validate,
    test_and_return,
)

__version__ = '0.1.1'

__package__ = 'disinfect'
__title__ = 'Disinfect'
__description__ = 'Disinfect: Destroy bad input.'
__uri__ = 'https://github.com/corverdevelopment/Disinfect/'

__author__ = 'Nils Corver'
__email__ = 'nils@corverdevelopment.nl'

__license__ = 'MIT'
__copyright__ = 'Copyright (c) 2016 Corver Development B.V.'

__all__ = [
    'Field',
    'Mapping',
    'MultiValueError',
    'Test',
    'validate',
    'test_and_return',
]
